# The MIT License (MIT).
#
# Copyright (c) 2024 Almaz Ilaletdinov <a.ilaletdinov@yandex.ru>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.

"""Ondivi app types."""

from typing import TypedDict

DiffStr = str
# Diff str is out of `git diff` command
# Example:
# +---------------------------------------------------------------------+
# │ diff --git a/ondivi/__main__.py b/ondivi/__main__.py                |
# | index 669d0ff..7a518fa 100644                                       |
# | --- a/ondivi/__main__.py                                            |
# | +++ b/ondivi/__main__.py                                            |
# | @@ -26,0 +27,2 @@ from git import Repo                              |
# | +Diff = str                                                         |
# | +FileName = str                                                     |
# | @@ -28 +30,2 @@ from git import Repo                                |
# | -def define_changed_lines(diff):                                    |
# | +                                                                   |
# | +def define_changed_lines(diff: Diff) -> dict[FileName, list[int]]: |
# +---------------------------------------------------------------------+

FileNameStr = str
# Filename with path to file
# This name must be equal in git diff and linter out

ViolationStr = str
# One line of violation output
# Example:
# "src/app_types/listable.py:23:1: UP035 Import from `collections.abc` instead: `Sequence`""

LinterAdditionalMessageStr = str
# Some additional messages from checkers which not contain code violation
# For example ruff write to stdout "Found 17 errors."
# This line not needed for ondivi, but we must send it to user in some cases
# See: "--only-violations" option

ActualViolationsListStr = list[str]
# List of violations filtered by ondivi

ViolationFormatStr = str
# Template for parsing linter messages. The template should include the following named parts:',
# {filename}   The name of the file with the error/warning',
# {line_num}   The line number with the error/warning (integer)',
# See "--format" option, https://github.com/r1chardj0n3s/parse

BaselineStr = str
# Branch name or commit hash

FromFilePathStr = str
# Path to file with violations in utf-8 encoding


class ParsedViolation(TypedDict):
    """Parsed violation.

    After parsing parse.parse takes dict type
    """

    filename: FileNameStr
    line_num: int



# The MIT License (MIT).
#
# Copyright (c) 2024 Almaz Ilaletdinov <a.ilaletdinov@yandex.ru>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.

"""Define changed lines in file."""

from ondivi._internal.types import DiffStr, FileNameStr  # noqa: WPS436. _internal allow into ondivi app


def define_changed_lines(diff: DiffStr) -> dict[FileNameStr, list[int]]:
    """Define changed lines in file.

    Example of diff:

    +---------------------------------------------------------------------+
    │ diff --git a/ondivi/__main__.py b/ondivi/__main__.py                | <- filename
    | index 669d0ff..7a518fa 100644                                       |
    | --- a/ondivi/__main__.py                                            |
    | +++ b/ondivi/__main__.py                                            |
    | @@ -26,0 +27,2 @@ from git import Repo                              | <- Changed lines = [27, 28]
    | +Diff = str                                                         |
    | +FileName = str                                                     |
    | @@ -28 +30,2 @@ from git import Repo                                | <- Changed lines = [27, 28, 30, 31]
    | -def define_changed_lines(diff):                                    |
    | +                                                                   |
    | +def define_changed_lines(diff: Diff) -> dict[FileName, list[int]]: |
    +---------------------------------------------------------------------+

    :param diff: DiffStr
    :return: dict[FileNameStr, list[int]]
    """
    changed_lines: dict[FileNameStr, list[int]] = {}
    current_file = ''
    for line in diff.splitlines():
        if _line_contain_filename(line):
            current_file = line.split(' b/')[-1].strip()
            changed_lines[current_file] = []
        elif _diff_line_contain_changed_lines(line):
            changed_lines[current_file].extend(_changed_lines(line))
    return changed_lines


def _line_contain_filename(diff_line: str) -> bool:
    return diff_line.startswith('diff --git')


def _diff_line_contain_changed_lines(diff_line: str) -> bool:
    return diff_line.startswith('@@')


def _changed_lines(diff_line: str) -> list[int]:
    """Changed lines.

    >>> _changed_lines('@@ -28 +30,2 @@ from git import Repo')
    [30, 31]

    :param diff_line: str
    :return: list[int]
    """
    splitted_line = diff_line.split('@@')[1].strip()
    added_lines = splitted_line.split('+')[1]
    start_line = int(
        added_lines.split(',')[0],
    )
    num_lines = 0
    if ',' in added_lines:
        num_lines = int(added_lines.split(',')[1]) - 1
    return list(range(
        start_line, start_line + num_lines + 1,
    ))



# The MIT License (MIT).
#
# Copyright (c) 2024 Almaz Ilaletdinov <a.ilaletdinov@yandex.ru>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.

"""Ondivi (Only diff violations).

Python script filtering coding violations, identified by static analysis,
only for changed lines in a Git repo.
"""

# flake8: noqa: WPS436. _internal allow into ondivi app

from __future__ import annotations

import sys
import traceback
from pathlib import Path

import click
from git import Repo
from git.exc import GitCommandError

from ondivi._internal.define_changed_lines import define_changed_lines
from ondivi._internal.filter_out_violations import filter_out_violations
from ondivi._internal.types import (
    ActualViolationsListStr,
    BaselineStr,
    DiffStr,
    FromFilePathStr,
    LinterAdditionalMessageStr,
    ViolationFormatStr,
    ViolationStr,
)


def controller(
    diff: DiffStr,
    linter_out: list[ViolationStr | LinterAdditionalMessageStr],
    violation_format: ViolationFormatStr,
    only_violations: bool,
) -> tuple[ActualViolationsListStr, bool]:
    """Entrypoint.

    :param diff: Diff
    :param linter_out: list[str]
    :param violation_format: ViolationFormatStr
    :param only_violations: bool
    :return: tuple[ActualViolationsListStr, bool]
    """
    return filter_out_violations(
        define_changed_lines(diff),
        linter_out,
        violation_format,
        only_violations,
    )


def _linter_output_from_file(file_path: FromFilePathStr) -> list[str]:
    if not Path(file_path).exists():
        sys.stdout.write('File with violations "{0}" not found\n'.format(file_path))
        sys.exit(1)
    return Path(file_path).read_text(encoding='utf-8').strip().splitlines()


def cli(
    baseline: BaselineStr,
    fromfile: FromFilePathStr | None,
    violation_format: ViolationFormatStr,
    only_violations: bool,
) -> None:
    """Controller with CLI side effects.

    :param baseline: BaselineStr
    :param fromfile: FromFilePathStr | None
    :param violation_format: ViolationFormatStr
    :param only_violations: bool
    """
    linter_output = _linter_output_from_file(fromfile) if fromfile else sys.stdin.read().strip().splitlines()
    try:
        diff = Repo('.').git.diff('--unified=0', baseline)
    except GitCommandError:
        sys.stdout.write('Revision "{0}" not found'.format(baseline))
        sys.exit(1)
    filtered_lines, violation_found = controller(
        diff,
        linter_output,
        violation_format,
        only_violations,
    )
    sys.stdout.write('\n'.join(filtered_lines))
    sys.stdout.write('\n')
    if violation_found:
        sys.exit(1)


# flake8: noqa: DAR101. Broke --help out
@click.command()
@click.option(
    '--baseline',
    default='master',
    help=' '.join([
        'Commit or branch which will contain legacy code.',
        'Program filter out violations on baseline',
        '(default: "master")',
    ]),
)
@click.option(
    '--fromfile',
    default=None,
    help='Path to file with violations. Expected "utf-8" encoding',
)
@click.option(
    '--format',
    'violation_format',
    default='{filename}:{line_num:d}{other}',
    help=''.join([
        'Template for parsing linter messages. The template should include the following named parts:\n\n',
        '{filename}   The name of the file with the error/warning\n',
        '{line_num}   The line number with the error/warning (integer)\n\n',
        'Example usage:\n\n',
        '--format "{filename}:{line_num:d}{other}"\n\n',
        'In this example, the linter message\n\n',
        '"src/app_types/listable.py:23:1: UP035 Import from collections.abc instead: Sequence"\n\n',
        'will be recognized and parsed into the following components:\n\n',
        ' - filename: "src/app_types/listable.py"\n\t\t\t\t',
        ' - line_num: 23\n\t\t\t\t\t',
        ' - other: :1: "UP035 Import from collections.abc instead: Sequence"\n\n',
        'Ensure that the template matches the format of the messages generated by your linter.\n\t\t\t\t',
        '(default: "{filename}:{line_num:d}{other}")',
    ]),
)
@click.option(
    '--only-violations',
    default=False,
    help='Show only violations',
    is_flag=True,
)
def main(baseline: str, fromfile: str | None, violation_format: str, only_violations: bool) -> None:
    """Ondivi (Only diff violations).

    Python script filtering coding violations, identified by static analysis,
    only for changed lines in a Git repo.
    Usage example:

    flake8 script.py | ondivi
    """
    try:
        cli(baseline, fromfile, violation_format, only_violations)
    except Exception as err:  # noqa: BLE001. Application entrypoint
        sys.stdout.write('\n'.join([
            'Ondivi fail with: "{0}"'.format(err),
            'Please submit it to https://github.com/blablatdinov/ondivi/issues',
            'Copy and paste this stack trace to GitHub:',
            '========================================',
            traceback.format_exc(),
        ]))
        sys.exit(1)
